package	jp.co.sds_corp.transfer;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import jp.co.sds_corp.transfer.filter.IFilter;
import jp.co.sds_corp.transfer.filter.IFilterable;
import jp.co.sds_corp.transfer.mixer.IMixer;
import jp.co.sds_corp.transfer.output.IOutput;
import jp.co.sds_corp.transfer.source.ISource;
import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.daemon.IService;
import jp.co.sds_corp.util.daemon.config.ConfigUtility;
import jp.co.sds_corp.util.daemon.core.AbstractDaemon;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

/**
データの取得／管理を行なうメインクラスです。
<P>
<PRE>
+---------------+          +---------+
|               |<-IFilter-| ISource |<-外部ソース
|               |          +---------+
|               |
|               |-IFilter->+---------+
| ServiceLoader |          | IMixer  |
|               |<---------+---------+
|               |
|               |          +---------+
|               |-IFilter->| IOutput |-ILayout->出力先
+---------------+          +---------+
</PRE>
ServiceLoaderがXMLを読み込んで、全てのインスタンスの生成と管理
初期化・処理開始等を制御します。

{@link ISource}から受け取ったデータは、一度{@link IMixer}に送られます。
{@link IMixer}では複数の${ISource}のデータを処理し
ServiceLoaderにデータを送り返します。
{@link IMixer}から送られたデータは、OutputとAdapterに送られ
指定された出力先に送信されます。
<P>
XMLは以下の形式です。
<PRE>
&lt;Daemon name="ServiceLoader" class="jp.co.sds-corp.co.jp.transfer.ServiceLoader"&gt;
</PRE>
<P>
Mixer要素は、{@link IMixer}クラスを指定します。<BR>
{@link IMixer}は、複数の{@link ISource}からデータ受信イベントを受け取り
ServiceLoaderにデータを返します。
<PRE>
&lt;Mixer name="Mixer" class="jp.co.sds_corp.co.jp.transfer.source.SimpleMixer"&gt;
&lt;/Mixer&gt;
</PRE>

<P>
Source要素は、データの取得元{@link ISource}を指定します。<BR>
子要素については個別のクラスの説明を参照してください。
<PRE>
&lt;Source name="Source" class="jp.co.sds_corp.co.jp.transfer.source.XXSource"&gt;
	&lt;url&gt;http://www.sds-corp.co.jp/vfx/servlet/ForwardFeed&lt;/url&gt;
	&lt;filter class="jp.co.sds_corp.transfer.filter.SanityFilter"/&gt;
	&lt;filter class="jp.co.sds_corp.transfer.filter.SanityFilter"&gt;
		&lt;ccypair&gt;USDJPY&lt;/ccypair&gt;
	&lt;/filter&gt;
&lt;/Source&gt;
</PRE>

<P>
Output要素は、データの出力先{@link IOutput}を指定します。<BR>
子要素については個別のクラスの説明を参照してください。
<PRE>
&lt;Output name="File" class="jp.co.sds-corp.co.jp.transfer.output.FileOutput"&gt;
	&lt;file&gt;data.hist&lt;/file&gt;
	&lt;format&gt;%C %c %T %B %A %U{yyyy/MM/dd HH:mm:ss}&lt;/format&gt;
&lt;/Output&gt;
&lt;Output name="Db" class="jp.co.sds_corp.co.jp.transfer.output.DbOutput"&gt;
	&lt;sql&gt;update table set bid=%b where ccy1=%C&lt;/sql&gt;
&lt;/Output&gt;
</PRE>
&lt;/Daemon&gt;
*/
public class ServiceLoader extends AbstractDaemon implements IReceiver
{
	static final CmnLogger logger = CmnLogFactory.getLogger();
	static ServiceLoader engine;
	String	config = "transfer.xml";
	HashMap<String,IService>	mapSrc = new HashMap<String,IService>();
	HashMap<String,IService>	mapOut = new HashMap<String,IService>();
	HashMap<String,IService>	mapMix = new HashMap<String,IService>();
	Vector<IService>	vecSvc = new Vector<IService>();
	ReceiverSupport support = new ReceiverSupport();
	HashMap<IDataKey,IData>	mapData = new HashMap<IDataKey,IData>();
	CheckerService	chkSvc = new CheckerService();

	/**
	* VM内のServiceLoaderインスタンスを返します。
	* @return	ServiceLoaderインスタンス
	*/
	public static ServiceLoader getInstance() {
		return engine;
	}

	/**
	* 指定された初期情報で、ServiceLoaderを初期化します。
	* @param	config	初期情報
	*/
	public void init( IConfiguration config ) throws Exception {
		engine = this;
		Map<String,? extends IService> map;
		map = createMixerMap( config );
		mapMix.putAll( map );
		vecSvc.addAll( map.values() );
		chkSvc.addService( map );

		map = createSourceMap( config );
		mapSrc.putAll( map );
		vecSvc.addAll( map.values() );
		chkSvc.addService( map );

		map = createOutputMap( config );
		mapOut.putAll( map );
		vecSvc.addAll( map.values() );
		chkSvc.addService( map );
		setService( chkSvc );
		createAllFilter( config );
	}

	/**
	* 指定された初期情報で、全ての{@link IMixer}を生成します。
	* @param	config	初期情報
	*/
	protected Map<String,IMixer> createMixerMap( IConfiguration config ) throws Exception {
		HashMap<String,IMixer> ret = new HashMap<String,IMixer>();
		Iterator<IConfiguration> itr = config.getChilds( "Mixer" );
		while ( itr.hasNext()) {
			IConfiguration c = (IConfiguration )itr.next();
			IMixer obj = createMixer( c );
			ret.put( c.getAttributes().get( "name" ), obj );
		}
		return ret;
	}

	/**
	* 指定された{@link IMixer}を生成します。
	* @param	config	初期情報
	*/
	protected IMixer createMixer( IConfiguration config )
		throws Exception
	{
		Object iObj = ConfigUtility.createInstance( config );
		IMixer obj = (IMixer)iObj;
		logger.debug( this, "loading mixer: " + obj.getClass().getName() );
		obj.setName( config.getAttributes().get( "name" ) );
		obj.init( config );
		addFilter( obj, config );
		obj.addReceiveListener( new IReceiveListener() {
			private static final long serialVersionUID = 1L;
			public void received( ReceiveEvent e ) throws Exception {
				recvMixerEvent( e );
			}
			public void statusChanged( StatusChangeEvent e ) throws Exception {
				recvMixerStatus( e );
			}
		} );

		return obj;
	}

	/**
	* 指定された初期情報で、全ての{@link ISource}を生成します。
	* @param	config	初期情報
	*/
	protected Map<String,ISource> createSourceMap( IConfiguration config ) throws Exception {
		HashMap<String,ISource> ret = new HashMap<String,ISource>();
		Iterator<IConfiguration> itr = config.getChilds( "Source" );
		while ( itr.hasNext()) {
			IConfiguration c = (IConfiguration )itr.next();
			ISource obj = createSource( c );
			ret.put( c.getAttributes().get( "name" ), obj );
		}
		return ret;
	}

	/**
	* 指定された{@link ISource}を生成します。
	* @param	config	初期情報
	*/
	protected ISource createSource( IConfiguration config )
		throws Exception
	{
		Object iObj = ConfigUtility.createInstance( config );
		ISource obj = (ISource)iObj;
		logger.debug( this, "loading source: " + obj.getClass().getName() );
		obj.setName( config.getAttributes().get( "name" ) );
		addFilter( obj, config );
		obj.addReceiveListener( new IReceiveListener() {
			private static final long serialVersionUID = 1L;
			public void received( ReceiveEvent e ) throws Exception {
				recvSourceEvent( e );
			}
			public void statusChanged( StatusChangeEvent e ) throws Exception {
				recvSourceStatus( e );
			}
		} );

		obj.init( config );
		return obj;
	}

	/**
	* 指定された{@link ISource}に{@link IFilter}を生成して設定します。
	* @param	obj		設定するソース
	* @param	config	初期情報
	*/
	protected void addFilter( IFilterable obj, IConfiguration config )
		throws Exception
	{
		Iterator<IConfiguration> itr = config.getChilds( "filter" );
		while ( itr.hasNext()) {
			IConfiguration c = (IConfiguration )itr.next();
			IFilter flt = (IFilter )ConfigUtility.createInstance( c );
			flt.init( c );
			obj.addFilter( flt );
		}
	}

	/**
	* 指定された初期情報で、全ての{@link IOutput}を生成します。
	* @param	config	初期情報
	*/
	protected Map<String,IOutput> createOutputMap( IConfiguration config ) throws Exception {
		HashMap<String,IOutput> ret = new HashMap<String,IOutput>();
		Iterator<IConfiguration> itr = config.getChilds( "Output" );
		while ( itr.hasNext()) {
			IConfiguration c = (IConfiguration )itr.next();
			IOutput obj = createOutput( c );
			ret.put( c.getAttributes().get( "name" ), obj );
		}
		return ret;
	}

	/**
	* 指定された{@link IOutput}を生成します。
	*/
	protected IOutput createOutput( IConfiguration config )
		throws Exception
	{
		Object iObj = ConfigUtility.createInstance( config );
		IOutput obj = (IOutput)iObj;
		logger.debug( this, "loading output: " + obj.getClass().getName() );
		obj.setName( config.getAttributes().get( "name" ) );
		addFilter( obj, config );
		obj.init( config );
		return obj;
	}

	/**
	* ServiceLoaderの処理を開始します。
	*/
	public void start() {
		start( mapMix );
		start( mapOut );
		start( mapSrc );
		setStatus( Status.ALIVE );

		logger.info( this, "ServiceLoader started." );
	}

	private void start( Map<String,IService> map ) {
		Iterator<String> itr;
		itr = map.keySet().iterator();
		while ( itr.hasNext()) {
			String name = (String )itr.next();
			IService svc = (IService )map.get( name );
			logger.debug( this, "start: " + svc.getName() + ":"
					+ svc.getClass().getName() );
			try {
				svc.start();
			} catch ( Throwable e ) {
				logger.fatal( this, "service [" + svc + "] cannot start. ", e );
			}
		}
	}

	/**
	* このクラスに属する全てのサービスを停止します。
	*/
	public void terminate() {
		terminate( mapSrc );
		terminate( mapOut );
		terminate( mapMix );
		setStatus( Status.DEAD, "Terminated." );
		logger.info( this, "ServiceLoader terminated." );
	}

	private void terminate( Map<String,IService> map ) {
		Iterator<String> itr;
		itr = map.keySet().iterator();
		while ( itr.hasNext()) {
			String name = (String )itr.next();
			IService svc = (IService )map.get( name );
			logger.debug( this, "stop:" + svc );
			svc.setInterrupt( true );
			svc.stop();
		}
	}

	/**
	* 全てのThreadの終了を待ちます。
	*/
	public void joinAll() {
	}

	/**
	* このデーモンの文字列表現
	* @return   このデーモンの文字列表現
	*/
	public String toString() {
		String ret = super.toString();

		Iterator<String> itr;
		itr = mapMix.keySet().iterator();
		while ( itr.hasNext()) {
			String name = (String )itr.next();
			IMixer mix = (IMixer )mapMix.get( name );
			ret += "\n\tMIX:" + mix.toString();
		}
		itr = mapSrc.keySet().iterator();
		while ( itr.hasNext()) {
			String name = (String )itr.next();
			ISource src = (ISource )mapSrc.get( name );
			ret += "\n\tSRC:" + src.toString();
		}
		itr = mapOut.keySet().iterator();
		while ( itr.hasNext()) {
			String name = (String )itr.next();
			IOutput out = (IOutput )mapOut.get( name );
			ret += "\n\tOUT:" + out.toString();
		}

		return ret;
	}

	/**
	* Sourceからデータ受信イベントを通知します。
	* @param	e		データ受信イベント
	*/
	protected void recvSourceEvent( ReceiveEvent e ) throws Exception {
		ISource src = (ISource )e.getSource();
		if ( !src.isValid( e )) {
			return;
		}

		Iterator<String> itr;
		itr = mapMix.keySet().iterator();
		while ( itr.hasNext()) {
			String name = (String )itr.next();
			IMixer mix = (IMixer )mapMix.get( name );
			if ( !mix.isValid( e )) {
				continue;
			}
			mix.received( e );
		}
	}

	/**
	* Sourceから状態変更イベントを通知します。
	* @param	e		状態変更イベント
	*/
	protected void recvSourceStatus( StatusChangeEvent e ) throws Exception {
		Iterator<String> itr;
		itr = mapMix.keySet().iterator();
		while ( itr.hasNext()) {
			String name = (String )itr.next();
			IMixer mix = (IMixer )mapMix.get( name );
			mix.statusChanged( e );
		}
	}

	/**
	* Mixerからデータ受信イベントを通知します。
	* @param	e		データ受信イベント
	*/
	protected void recvMixerEvent( ReceiveEvent e ) throws Exception {
		mapData.put( e.getData().getKey(), e.getData() );

		fireReceive( e );

		Iterator<String> itr = mapOut.keySet().iterator();
		while ( itr.hasNext()) {
			String name = (String )itr.next();
			IOutput out = (IOutput )mapOut.get( name );
			if ( !out.isValid( e )) {
				continue;
			}
			try {
				out.received( e );
			} catch ( Throwable t ) {
				logger.error( this, "output error", t );
			}
		}
	}

	/**
	* Mixerから状態変更イベントを通知します。
	* @param	e		状態変更イベント
	*/
	protected void recvMixerStatus( StatusChangeEvent e ) throws Exception {
		Iterator<String> itr;
		itr = mapOut.keySet().iterator();
		while ( itr.hasNext()) {
			String name = (String )itr.next();
			IOutput out = (IOutput )mapOut.get( name );
			out.statusChanged( e );
		}
	}
	/**
	* 現在のデータを返します。
	*/
	public Collection<IData> getCurrent( IDataKey key ) {
		Vector<IData> ret = new Vector<IData>();
		Iterator<IDataKey> itr = mapData.keySet().iterator();
		while ( itr.hasNext() ) {
			IDataKey k = (IDataKey )itr.next();
			if ( key.match( k )) {
				ret.add( mapData.get( k ));
			}
		}
		return ret;
	}

	/**
	* データ受信を受け取るオブジェクトを追加します。
	* @param	l		データ受信を受け取るオブジェクト
	*/
	public void addReceiveListener( IReceiveListener l ) {
		support.addReceiveListener( l );
	}
	/**
	* データ受信を受け取るオブジェクトを削除します。
	* @param	l		データ受信を受け取るオブジェクト
	*/
	public void removeReceiveListener( IReceiveListener l ) {
		support.removeReceiveListener( l );
	}
	/**
	* データ受信を受け取るオブジェクトを追加します。
	* @param	key		対象のキー
	* @param	l		データ受信を受け取るオブジェクト
	*/
	public void addReceiveListener( IDataKey key, IReceiveListener l ) {
		support.addReceiveListener( key, l );
	}
	/**
	* データ受信を受け取るオブジェクトを削除します。
	* @param	key		対象のキー
	* @param	l		データ受信を受け取るオブジェクト
	*/
	public void removeReceiveListener( IDataKey key, IReceiveListener l ) {
		support.removeReceiveListener( key, l );
	}
	/**
	* データ受信イベントを通知します。
	* @param	e		データ受信イベント
	*/
	public void fireReceive( ReceiveEvent e ) throws Exception {
		support.fireReceive( e );
	}
	/**
	* 状態変更イベントを通知します。
	* @param	e		状態変更イベント
	*/
	public void fireStatusChange( StatusChangeEvent e ) throws Exception {
		support.fireStatusChange( e );
	}

}
